import os
import torch


def load_install_torchsde():
    """Try importing torchsde. If not installed, it will be installed and imported."""
    try:
        import torchsde
    except:
        os.system("pip install torchsde")
        import torchsde

    return torchsde


def load_install_torchdiffeq():
    """Try importing torchdiffeq. If not installed, it will be installed and imported."""
    try:
        import torchdiffeq
    except:
        os.system("pip install torchdiffeq")
        import torchdiffeq

    return torchdiffeq


def test_odeint(
    ODE, y0=torch.randn([200, 50]), t=torch.Tensor([0, 0.001, 0.002]), **kwargs
):
    torchdiffeq = load_install_torchdiffeq()

    """Generate some random data and test the ODE as it's passed through torchdiffeq.odeint."""
    return torchdiffeq.odeint(ODE, y0, t, **kwargs)


def test_sdeint(
    SDE, y0=torch.randn([200, 50]), t=torch.Tensor([0, 0.5, 1.0]), **kwargs
):
    torchsde = load_install_torchsde()
    """Generate some random data and test the ODE as it's passed through torchsde.sdeint."""
    return torchsde.sdeint(SDE, y0, t, dt=0.1, **kwargs)



# -- import packages: ----------------------------------------------------------
import torch
import ABCParse


# -- import local dependencies: ------------------------------------------------
from .core._base_neural_ode import BaseODE


# -- import standard libraries and define types: -------------------------------
from typing import Union, List, Any
NoneType = type(None)


# -- Main operational class: ---------------------------------------------------
class NeuralODE(BaseODE):
    DIFFEQ_TYPE = "ODE"
    def __init__(
        self,
        state_size: int,
        dt: float = 0.1,
        coef_diff: float = 0,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type="ito",
        noise_type="general",
    ):
        sigma_hidden = []
        sigma_output_bias = False
        brownian_dim = 1

        super().__init__()

        self.__config__(locals())

    def drift(self, y) -> torch.Tensor:
        return self.mu(y)

    def forward(self, t, y):
        return self.mu(y)




# -- import packages: ----------------------------------------------------------
import torch
import ABCParse


# -- import local dependencies: ------------------------------------------------
from . import core


# -- import standard libraries and define types: -------------------------------
from typing import Union, List, Any
NoneType = type(None)


# -- Main operational class: ---------------------------------------------------
class LatentPotentialSDE(core.BaseLatentSDE):
    DIFFEQ_TYPE = "SDE"

    def __init__(
        self,
        state_size,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        sigma_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        sigma_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type="ito",
        noise_type="general",
        brownian_dim=1,
        coef_drift: float = 1.,
        coef_diffusion: float = 1.,
        coef_prior_drift: float = 1.,
    ):
        super().__init__()

        # explicitly state that we are not learning a potential function by default
        mu_potential = False
        self.__config__(locals())
        self.potential = core.Potential(state_size)

    def drift(self, y) -> torch.Tensor:
        return self.mu(y) * self.coef_drift

    def prior_drift(self, y) -> torch.Tensor:
        return self.potential.potential_gradient(y) * self.coef_prior_drift

    def diffusion(self, y) -> torch.Tensor:
        return self.sigma(y).view(y.shape[0], y.shape[1], self.brownian_dim) * self.coef_diffusion




import ABCParse
import torch_nets


class DiffEqConfig(ABCParse.ABCParse):
    def __init__(
        self,
        state_size,
        mu_hidden=[2000, 2000],
        sigma_hidden=[400, 400],
        brownian_dim=1,
        mu_activation=["LeakyReLU"],
        sigma_activation=["LeakyReLU"],
        mu_dropout=[0.2],
        sigma_dropout=[0.2],
        mu_n_augment=0,
        sigma_n_augment=0,
        mu_bias=[True],
        sigma_bias=[True],
        mu_potential=False,
        sigma_potential=False,
        mu_output_bias=False,
        sigma_output_bias=False,
    ):
        self.__parse__(locals(), public = [None])

    @property
    def mu_in_features(self):
        return self._state_size

    @property
    def sigma_in_features(self):
        return self._state_size

    @property
    def mu_out_features(self):
        if self._mu_potential:
            return 1
        return self._state_size

    @property
    def sigma_out_features(self):
        if self._sigma_potential:
            return 1
        return self._state_size * self._brownian_dim

    @property
    def mu_output_bias(self):
        if self._mu_potential:
            return False
        return self._mu_output_bias

    @property
    def sigma_output_bias(self):
        if self._sigma_potential:
            return False
        return self._sigma_output_bias

    @property
    def mu_net_cls(self):
        if self._mu_n_augment:
            self._mu_kwargs = {"n_augment": self._mu_n_augment}
            return torch_nets.AugmentedTorchNet
        self._mu_kwargs = {}
        return torch_nets.TorchNet

    @property
    def sigma_net_cls(self):
        if self._sigma_n_augment:
            self._sigma_kwargs = {"n_augment": self._sigma_n_augment}
            return torch_nets.AugmentedTorchNet
        self._sigma_kwargs = {}
        return torch_nets.TorchNet

    @property
    def mu(self):
        return self.mu_net_cls(
            in_features=self.mu_in_features,
            out_features=self.mu_out_features,
            hidden=self._mu_hidden,
            activation=self._mu_activation,
            dropout=self._mu_dropout,
            bias=self._mu_bias,
            output_bias=self.mu_output_bias,
            **self._mu_kwargs,
        )

    @property
    def sigma(self):
        return self.sigma_net_cls(
            in_features=self.sigma_in_features,
            out_features=self.sigma_out_features,
            hidden=self._sigma_hidden,
            activation=self._sigma_activation,
            dropout=self._sigma_dropout,
            bias=self._sigma_bias,
            output_bias=self.sigma_output_bias,
            **self._sigma_kwargs
        )
    def __call__(self):
        return self.mu, self.sigma




import torch
import ABCParse
from abc import abstractmethod


# from ._base_neural_diffeq import BaseDiffEq
from ._base_neural_ode import BaseODE
from ._diffeq_config import DiffEqConfig


class BaseLatentODE(BaseODE):
    DIFFEQ_TYPE = "ODE"

    def __init__(self, *args, **kwargs):
        super().__init__()

        """
        Must call self.__config__(locals()) in the __init__ of theinheriting
        class.
        
        """

#     def __config__(self, kwargs):
#         """Sets up mu and sigma given params"""

#         self.__parse__(kwargs=kwargs)

#         self._config_kwargs = ABCParse.function_kwargs(func=DiffEqConfig, kwargs=kwargs)
#         configs = DiffEqConfig(**self._config_kwargs)
#         self.mu = configs.mu

    # -- required methods in child classes: ------------------------------------
    @abstractmethod
    def drift(self):
        """Called by self.f"""
        ...
    
#     def diffusion(self, y):
#         # keep for compatibility with torchsde.sdeint
#         """Called by self.g"""
#         return torch.zeros([y.shape[0], y.shape[1], self.brownian_dim])
    
    @abstractmethod
    def prior_drift(self):
        """Called by self.h"""
        ...

    def h(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Should return the output of self.diffusion"""
        return self.prior_drift(y)




# -- import packages: ----------------------------------------------------------
import torch


# -- operational class: --------------------------------------------------------
class Potential(torch.nn.Sequential):

    """Linear transform state of arbitrary dimension to a 1-D potential value"""

    def __init__(self, state_size: int):
        """
        Parameters:
        -----------
        state_size
            type: int

        Returns:
        --------
        None, instantiates class.
        """
        super().__init__()
        self.add_module("psi", torch.nn.Linear(state_size, 1, bias=False))

    def gradient(self, ψ, y):
        """use-case: output is directly psi (from a potential network)"""
        return torch.autograd.grad(ψ, y, torch.ones_like(ψ), create_graph=True)[0]

    def potential_gradient(self, y):
        """in this use-case, y is likely the output of a neural network"""
        y = y.requires_grad_()
        return self.gradient(self.psi(y), y)
    
    def __call__(self, y:torch.Tensor)->torch.Tensor:
        return self.psi(y.requires_grad_())



import torch
import ABCParse
from abc import abstractmethod

from typing import List

from ._base_neural_diffeq import BaseDiffEq
from ._diffeq_config import DiffEqConfig

class BaseODE(BaseDiffEq):
    DIFFEQ_TYPE = "ODE"

    def __init__(self, *args, **kwargs):
        super().__init__()

        """
        In inherited class, must pass:
            "dt"
            "coef_diff"
        
        Must call self.__config__(locals()) in the __init__
        of the inheriting class.        
        """

    def __config__(self, kwargs, private: List[str] = ["coef_diff", "dt"]):
        """Sets up mu and sigma given params"""

        self.__parse__(kwargs=kwargs, private=private, public = ['noise_type', 'sde_type'])

        self._config_kwargs = ABCParse.function_kwargs(func=DiffEqConfig, kwargs=kwargs)
        configs = DiffEqConfig(**self._config_kwargs)
        self.mu = configs.mu
        
    @property
    def _sqrt_dt(self):
        """Must pass `dt`"""
        return torch.sqrt(torch.Tensor([self._dt]).to(self.device))

    def _sigma(self, y):
        return torch.randn([y.shape[0], y.shape[1], self._brownian_dim], device=self.device)
    
    @property
    def device(self):
        return list(self.parameters())[0].device

    # -- required methods in child classes: ------------------------------------
    @abstractmethod
    def drift(self):
        """Called by self.f and/or self.forward"""
        ...
        
    def forward(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.drift(y)

    def diffusion(
        self, y: torch.Tensor, coef_diff: float = 0., dt: float = 0.1
    ) -> torch.Tensor:
        # keep for compatibility with torchsde.sdeint
        """Called by self.g"""
        return  self._sigma(y) * self._coef_diff * self._sqrt_dt




import torch
import ABCParse
from abc import abstractmethod


from ._base_neural_diffeq import BaseDiffEq
from ._diffeq_config import DiffEqConfig


class BaseLatentSDE(BaseDiffEq):
    DIFFEQ_TYPE = "SDE"

    def __init__(self, *args, **kwargs):
        super().__init__()

        """
        Must call self.__config__(locals()) in the __init__ of theinheriting
        class.
        
        
        """

    def __config__(self, kwargs):
        """Sets up mu and sigma given params"""

        self.__parse__(kwargs=kwargs, public = ['noise_type', 'sde_type'])

        self._config_kwargs = ABCParse.function_kwargs(func=DiffEqConfig, kwargs=kwargs)
        configs = DiffEqConfig(**self._config_kwargs)

        self.mu = configs.mu
        self.sigma = configs.sigma

    # -- required methods in child classes: ------------------------------------
    @abstractmethod
    def drift(self):
        """Called by self.f"""
        ...

    @abstractmethod
    def diffusion(self):
        """Called by self.g"""
        ...
        
    @abstractmethod
    def prior_drift(self):
        """Called by self.h"""
        ...
        
    def h(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Should return the output of self.diffusion"""
        return self.prior_drift(y)




__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["mvinyard.ai@gmail.com"])
__version__ = "0.3.2"


# -- import subpackages: -------------------------------------------------------
from . import core


# -- import modules: -----------------------------------------------------------
from ._neural_ode import NeuralODE
from ._neural_sde import NeuralSDE
from ._potential_sde import PotentialSDE
from ._potential_ode import PotentialODE
from ._latent_potential_ode import LatentPotentialODE
from ._latent_potential_sde import LatentPotentialSDE





# -- import packages: ----------------------------------------------------------
import torch
import ABCParse


# -- import standard libraries: ------------------------------------------------
from abc import abstractmethod


# -- import local dependencies: ------------------------------------------------
from ._diffeq_config import DiffEqConfig


# -- Main operational class: ---------------------------------------------------
class BaseDiffEq(torch.nn.Module, ABCParse.ABCParse):
    DIFFEQ_TYPE = ""
    def __init__(self, *args, **kwargs):
        super().__init__()
        
        """
        Must call self.__config__(locals()) in the __init__ of the inheriting
        class.        
        """

    # -- required methods in child classes: ------------------------------------
    @abstractmethod
    def drift(self):
        """Called by self.f"""
        ...

    @abstractmethod
    def diffusion(self):
        """Called by self.g"""
        ...

    def f(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Should return the output of self.drift"""
        return self.drift(y)

    def g(self, t: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Should return the output of self.diffusion"""
        return self.diffusion(y)



# -- import packages: ----------------------------------------------------------
import torch
import ABCParse


# -- import local dependencies: ------------------------------------------------
from .core import BaseODE


# -- import standard libraries and define types: -------------------------------
from typing import Union, List, Any

NoneType = type(None)


# -- Main operational class: ---------------------------------------------------
class PotentialODE(BaseODE):
    DIFFEQ_TYPE = "ODE"
    sde_type = "ito"
    noise_type = "general"
    _brownian_dim = 1
        
    def __init__(
        self,
        state_size,
        dt: float = 0.1,
        coef_diff: float = 0,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
    ):
        super().__init__()
        
        mu_potential = True

        self.__config__(locals())

    def _potential(self, y):
        return self.mu(y)

    def _gradient(self, ψ, y):
        """use-case: output is directly psi (from a potential network)"""
        return torch.autograd.grad(ψ, y, torch.ones_like(ψ), create_graph=True)[0]

    def drift(self, y) -> torch.Tensor:
        y = y.requires_grad_()
        ψ = self._potential(y)
        return self._gradient(ψ, y)



# from . import base_models

from ._potential import Potential
from ._diffeq_config import DiffEqConfig
from ._base_neural_diffeq import BaseDiffEq
from ._base_neural_ode import BaseODE
from ._base_neural_sde import BaseSDE
from ._base_latent_ode import BaseLatentODE
from ._base_latent_sde import BaseLatentSDE